using Adaptation.Eaf.Management.ConfigurationData.CellAutomation;
using Adaptation.Ifx.Eaf.EquipmentConnector.File.Configuration;
using Adaptation.Shared;
using Adaptation.Shared.Duplicator;
using Adaptation.Shared.Methods;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text.Json;
using System.Threading;

namespace Adaptation.FileHandlers.Priority;

public class FileRead : Shared.FileRead, IFileRead
{

    private readonly Timer _Timer;

    public FileRead(ISMTP smtp, Dictionary<string, string> fileParameter, string cellInstanceName, int? connectionCount, string cellInstanceConnectionName, FileConnectorConfiguration fileConnectorConfiguration, string equipmentTypeName, string parameterizedModelObjectDefinitionType, IList<ModelObjectParameterDefinition> modelObjectParameters, string equipmentDictionaryName, Dictionary<string, List<long>> dummyRuns, Dictionary<long, List<string>> staticRuns, bool useCyclicalForDescription, bool isEAFHosted) :
        base(new Description(), false, smtp, fileParameter, cellInstanceName, connectionCount, cellInstanceConnectionName, fileConnectorConfiguration, equipmentTypeName, parameterizedModelObjectDefinitionType, modelObjectParameters, equipmentDictionaryName, dummyRuns, staticRuns, useCyclicalForDescription, isEAFHosted: connectionCount is null)
    {
        _MinFileLength = 10;
        _NullData = string.Empty;
        _Logistics = new(this);
        if (_FileParameter is null)
            throw new Exception(cellInstanceConnectionName);
        if (_ModelObjectParameterDefinitions is null)
            throw new Exception(cellInstanceConnectionName);
        if (!_IsDuplicator)
            throw new Exception(cellInstanceConnectionName);
        if (_IsEAFHosted)
            NestExistingFiles(_FileConnectorConfiguration);
        if (!Debugger.IsAttached && fileConnectorConfiguration.PreProcessingMode != FileConnectorConfiguration.PreProcessingModeEnum.Process)
            _Timer = new Timer(Callback, null, (int)(fileConnectorConfiguration.FileScanningIntervalInSeconds * 1000), Timeout.Infinite);
        else
        {
            _Timer = new Timer(Callback, null, Timeout.Infinite, Timeout.Infinite);
            Callback(null);
        }
    }

    void IFileRead.Move(Tuple<string, Test[], JsonElement[], List<FileInfo>> extractResults, Exception exception) => Move(extractResults);

    void IFileRead.WaitForThread() => WaitForThread(thread: null, threadExceptions: null);

    string IFileRead.GetEventDescription()
    {
        string result = _Description.GetEventDescription();
        return result;
    }

    List<string> IFileRead.GetHeaderNames()
    {
        List<string> results = _Description.GetHeaderNames();
        return results;
    }

    string[] IFileRead.Move(Tuple<string, Test[], JsonElement[], List<FileInfo>> extractResults, string to, string from, string resolvedFileLocation, Exception exception)
    {
        string[] results = Move(extractResults, to, from, resolvedFileLocation, exception);
        return results;
    }

    JsonProperty[] IFileRead.GetDefault()
    {
        JsonProperty[] results = _Description.GetDefault(this, _Logistics);
        return results;
    }

    Dictionary<string, string> IFileRead.GetDisplayNamesJsonElement()
    {
        Dictionary<string, string> results = _Description.GetDisplayNamesJsonElement(this);
        return results;
    }

    List<IDescription> IFileRead.GetDescriptions(IFileRead fileRead, List<Test> tests, IProcessData processData)
    {
        List<IDescription> results = _Description.GetDescriptions(fileRead, _Logistics, tests, processData);
        return results;
    }

    Tuple<string, Test[], JsonElement[], List<FileInfo>> IFileRead.GetExtractResult(string reportFullPath, string eventName)
    {
        Tuple<string, Test[], JsonElement[], List<FileInfo>> results;
        if (string.IsNullOrEmpty(eventName))
            throw new Exception();
        _ReportFullPath = reportFullPath;
        DateTime dateTime = DateTime.Now;
        results = GetExtractResult(reportFullPath, dateTime);
        if (results.Item3 is null)
            results = new Tuple<string, Test[], JsonElement[], List<FileInfo>>(results.Item1, Array.Empty<Test>(), JsonSerializer.Deserialize<JsonElement[]>("[]"), results.Item4);
        if (results.Item3.Length > 0 && _IsEAFHosted)
            WritePDSF(this, results.Item3);
        UpdateLastTicksDuration(DateTime.Now.Ticks - dateTime.Ticks);
        return results;
    }

    Tuple<string, Test[], JsonElement[], List<FileInfo>> IFileRead.ReExtract()
    {
        Tuple<string, Test[], JsonElement[], List<FileInfo>> results;
        List<string> headerNames = _Description.GetHeaderNames();
        Dictionary<string, string> keyValuePairs = _Description.GetDisplayNamesJsonElement(this);
        results = ReExtract(this, headerNames, keyValuePairs);
        return results;
    }

#nullable enable

    private static ReadOnlyCollection<Record> GetRecords(string directory, string searchPattern)
    {
        List<Record> results = new();
        string text;
        Record? record;
        string[] files;
        List<Record>? collection;
        Dictionary<string, List<Record>> keyValuePairs = new();
        string[] directories = Directory.GetDirectories(directory, "*", SearchOption.TopDirectoryOnly);
        foreach (string subDirectory in directories)
        {
            keyValuePairs.Clear();
            files = Directory.GetFiles(subDirectory, searchPattern, SearchOption.TopDirectoryOnly);
            foreach (string file in files)
            {
                text = File.ReadAllText(file);
                if (string.IsNullOrEmpty(text) || text[0] == '[')
                    continue;
                record = JsonSerializer.Deserialize<Record>(text);
                if (record is null || record.Id == 0)
                    continue;
                if (!keyValuePairs.TryGetValue(record.RemoteIpAddress, out collection))
                {
                    keyValuePairs.Add(record.RemoteIpAddress, new());
                    if (!keyValuePairs.TryGetValue(record.RemoteIpAddress, out collection))
                        throw new Exception();
                }
                collection.Add(record);
            }
            foreach (KeyValuePair<string, List<Record>> keyValuePair in keyValuePairs)
            {
                if (keyValuePair.Value.Count == 1)
                    results.Add(keyValuePair.Value[0]);
                else
                {
                    record = keyValuePair.Value.Select(record => new KeyValuePair<long, Record>(record.Time, record)).OrderBy(pair => pair.Key).Last().Value;
                    results.Add(record);
                }
            }
        }
        return new(results);
    }

    private static int? GetInverse(int value) =>
        value switch
        {
            1 => 3,
            2 => 2,
            3 => 1,
            _ => null
        };

    private static int? GetInverse(double value)
    {
        int? result;
        if (value > 3)
            result = null;
        else if (value > 2)
            result = 1;
        else if (value > 1)
            result = 2;
        else if (value > 0)
            result = 3;
        else
            result = null;
        return result;
    }

    private static ReadOnlyDictionary<int, Aggregation> GetKeyValuePairs(Dictionary<int, List<Record>> keyValuePairs)
    {
        Dictionary<int, Aggregation> results = new();
        Aggregation aggregation;
        int? inverse;
        double average;
        List<int> collection = new();
        foreach (KeyValuePair<int, List<Record>> keyValuePair in keyValuePairs)
        {
            collection.Clear();
            foreach (Record record in keyValuePair.Value)
            {
                inverse = GetInverse(record.Value);
                if (inverse is null)
                    continue;
                collection.Add(inverse.Value);
            }
            average = collection.Average();
            inverse = GetInverse(average);
            aggregation = new(average.ToString("0.000"),
                              keyValuePair.Value.Count,
                              inverse,
                              keyValuePair.Value.Max(record => record.Value),
                              keyValuePair.Value.Min(record => record.Value),
                              new(keyValuePair.Value),
                              keyValuePair.Value.Sum(record => record.Value));
            results.Add(keyValuePair.Key, aggregation);
        }
        return new(results);
    }

    private static ReadOnlyDictionary<int, Aggregation> GetKeyValuePairs(string directory, string searchPattern)
    {
        ReadOnlyDictionary<int, Aggregation> results;
        List<Record>? collection;
        Dictionary<int, List<Record>> keyValuePairs = new();
        ReadOnlyCollection<Record> records = GetRecords(directory, searchPattern);
        foreach (Record record in records)
        {
            if (!keyValuePairs.TryGetValue(record.Id, out collection))
            {
                keyValuePairs.Add(record.Id, new());
                if (!keyValuePairs.TryGetValue(record.Id, out collection))
                    throw new Exception();
            }
            collection.Add(record);
        }
        results = GetKeyValuePairs(keyValuePairs);
        return results;
    }

    private static void WriteFiles(string sourceFileLocation, string sourceFileFilter, string targetFileLocation)
    {
        string json;
        string jsonFile;
        string directoryName;
        if (!Directory.Exists(sourceFileLocation))
            _ = Directory.CreateDirectory(sourceFileLocation);
        if (!Directory.Exists(targetFileLocation))
            _ = Directory.CreateDirectory(targetFileLocation);
        ReadOnlyDictionary<int, Aggregation> keyValuePairs;
        JsonSerializerOptions jsonSerializerOptions = new() { WriteIndented = true };
        string[] directories = Directory.GetDirectories(sourceFileLocation, "*", SearchOption.TopDirectoryOnly);
        foreach (string directory in directories)
        {
            directoryName = Path.GetFileName(directory);
            keyValuePairs = GetKeyValuePairs(directory, sourceFileFilter);
            jsonFile = Path.Combine(targetFileLocation, $"{directoryName}.json");
            json = JsonSerializer.Serialize(keyValuePairs, jsonSerializerOptions);
            File.WriteAllText(jsonFile, json);
        }
    }

    private void Callback(object state)
    {
        try
        {
            if (_IsEAFHosted)
                WriteFiles(_FileConnectorConfiguration.SourceFileLocation, _FileConnectorConfiguration.SourceFileFilter, _FileConnectorConfiguration.TargetFileLocation);
        }
        catch (Exception exception)
        {
            string subject = string.Concat("Exception:", _CellInstanceConnectionName);
            string body = string.Concat(exception.Message, Environment.NewLine, Environment.NewLine, exception.StackTrace);
            try
            { _SMTP.SendHighPriorityEmailMessage(subject, body); }
            catch (Exception) { }
        }
        try
        {
            if (_FileConnectorConfiguration?.FileScanningIntervalInSeconds is null)
                throw new Exception();
            TimeSpan timeSpan = new(DateTime.Now.AddSeconds(_FileConnectorConfiguration.FileScanningIntervalInSeconds.Value).Ticks - DateTime.Now.Ticks);
            _ = _Timer.Change((long)timeSpan.TotalMilliseconds, Timeout.Infinite);
        }
        catch (Exception exception)
        {
            string subject = string.Concat("Exception:", _CellInstanceConnectionName);
            string body = string.Concat(exception.Message, Environment.NewLine, Environment.NewLine, exception.StackTrace);
            try
            { _SMTP.SendHighPriorityEmailMessage(subject, body); }
            catch (Exception) { }
        }
    }

#pragma warning disable IDE0060
    private Tuple<string, Test[], JsonElement[], List<FileInfo>> GetExtractResult(string reportFullPath, DateTime dateTime)
#pragma warning restore IDE0060
    {
        Tuple<string, Test[], JsonElement[], List<FileInfo>> results;
        _Logistics = new Logistics(reportFullPath, $"LOGISTICS_1{'\t'}A_JOBID={"BACKLOG"};A_MES_ENTITY={"BACKLOG"};");
        results = new(_Logistics.Logistics1[0], Array.Empty<Test>(), Array.Empty<JsonElement>(), new List<FileInfo>());
        return results;
    }

}